/*

  PONGDuino : a Pong game for Arduino and 16x2 LCD display

  Copyright (c) 2013, Xavier Miller (xavier.miller@cauwe.org)
  http://www.xaviermiller.be

  All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:

  - Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.

*/

#include "pong.h"

Pong::Pong(IPongReceiver *receiver, int width, int height, int paddle_size, int paddle_width, int ball_size, int dir_col, int dir_line, int score_max)
    :
      m_receiver(receiver),
      m_width(width),
      m_height(height),
      m_col(0),
      m_line(0),
      m_paddle_size(paddle_size),
      m_paddle_width(paddle_width),
      m_ball_size(ball_size),
      m_dir_col(dir_col),
      m_dir_line(dir_line),
      m_left_paddle(0),
      m_right_paddle(0),
      m_score1(0),
      m_score2(0),
      m_score_max(score_max),
      m_status(Paused),
      m_pause_type(PauseSplash)
{
    Reset();
}

void Pong::Reset()
{
    m_score1 = 0;
    m_score2 = 0;
    m_pause_type = Pong::PauseSplash;
    CenterPaddles();
    NewBall();
}

void Pong::CenterPaddles()
{
    m_left_paddle = (m_height - m_paddle_size) >> 1;
    m_right_paddle = m_left_paddle;
}

void Pong::CenterBall()
{
    m_col = m_width >> 1;
    m_line = m_height >> 1;
}

void Pong::NewBall()
{
    CenterBall();

    m_status = Pong::Paused;
    m_receiver->Pause();

    Draw();
}

void Pong::KeyPause()
{
    if (m_status == Pong::Paused)
    {
        m_status = Pong::Running;
        m_receiver->Continue();
    }
    else
    {
        m_status = Pong::Paused;
        m_pause_type = Pong::PauseDefault;
        m_receiver->Pause();
    }
    Draw();
}

void Pong::KeyUp(PongPlayer player)
{
    if (m_status == Pong::Running)
    {
        MovePlayer(player, Pong::Up);
    }
}

void Pong::KeyDown(PongPlayer player)
{
    if (m_status == Pong::Running)
    {
        MovePlayer(player, Pong::Down);
    }
}

void Pong::TimerTick()
{
    if (m_status == Pong::Paused)
    {
        m_receiver->Pause();
        return;
    }

    m_col += m_dir_col;
    // left wall collision
    if (m_col < m_paddle_width)
    {
        m_dir_col = -m_dir_col;
        if ((m_left_paddle <= m_line) && (m_line < m_left_paddle+m_paddle_size))
        {
            m_receiver->BeepPaddle();
            m_col += 2 * m_dir_col;
        }
        else
        {
            m_receiver->BeepLost();
            m_score2++;
            m_pause_type = Pong::PausePlayer2Win;
            NewBall();
            return;
        }
    }
    // right wall collision
    if (m_col + m_ball_size > m_width - m_paddle_width)
    {
        m_dir_col = -m_dir_col;
        if ((m_right_paddle <= m_line) && (m_line < m_right_paddle + m_paddle_size))
        {
            m_receiver->BeepPaddle();
            m_col += 2 * m_dir_col;
        }
        else
        {
            m_receiver->BeepLost();
            m_score1++;
            m_pause_type = Pong::PausePlayer1Win;
            NewBall();
            return;
        }
    }

    // upper and lower walls collision
    m_line += m_dir_line;
    if ((m_line < 0) || (m_line + m_ball_size > m_height))
    {
        m_receiver->BeepWall();
        m_dir_line = -m_dir_line;
        m_line += 2 * m_dir_line;
    }

    Draw();
}

void Pong::MovePlayer(PongPlayer player, PongPadDirection direction)
{
    if (direction == Up)
    {
        if (player == Player1)
        {
            if (m_left_paddle)
                m_left_paddle--;
        }
        else
        {
            if (m_right_paddle)
                m_right_paddle--;
        }
    }
    else
    {
        if (player == Player1)
        {
            if (m_left_paddle < (m_height - m_paddle_size))
                m_left_paddle++;
        }
        else
        {
            if (m_right_paddle < (m_height - m_paddle_size))
                m_right_paddle++;
        }

    }
}

void Pong::Draw()
{
    if (m_status == Pong::Running)
    {
        m_receiver->GameScreen(m_score1, m_score2,
                               m_col, m_line, m_ball_size,
                               0, m_left_paddle,
                               m_width - m_paddle_width, m_right_paddle,
                               m_paddle_width,
                               m_paddle_size);
    }
    else
    {   const char *msg = "      ****      ";
        if ((m_score1 == m_score_max) || (m_score2 == m_score_max))
            m_pause_type = GameOver;

        switch (m_pause_type)
        {
        case PauseSplash:
            msg = "   PongDuino!   ";
            break;
        case PauseDefault:
            msg = "     Pause      ";
            break;
        case PausePlayer1Win:
            msg = " Player 1 Wins! ";
            break;
        case PausePlayer2Win:
            msg = " Player 2 Wins! ";
            break;
        case GameOver:
            msg = "G A M E  O V E R";
            break;
        }

        m_receiver->PauseScreen(m_score1, m_score2, (const unsigned char*) msg);

        if (m_pause_type == GameOver)
        {
            m_score1 = 0;
            m_score2 = 0;
            CenterPaddles();
            CenterBall();
        }
    }
}
